// On-Line Discovery of Dense Areas in Spatio-temporal Databases
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

#include "global.h"

unsigned long h(unsigned int id, unsigned long m, double hashMod)
{
	double d = (double) id * hashMod;
	return (unsigned long) floor(((double) m) * (d - floor(d)));
}

unsigned long h2(double id, unsigned long m, double hashMod)
{
	return (unsigned long) floor(id * hashMod * m);
}

int main(int argc, char** argv)
{
	char dataFile[1000] = "d";
	int gridsize = 250;
	unsigned int horizon = 20;
	unsigned int density = 1000;
	unsigned int numberOfStages = 2;
	unsigned int numberOfCounters = 2000;
	bool sampling = false;
	double samplingFactor = 0.01;
	int queryType = SNAPSHOT;

	time_t queryTime = 0;
	time_t updateTime = 0;

	for (int i = 1; i < argc; i++)
	{
		if (! strcmp(argv[i], "-df"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing data file." << endl;
				return -1;
			}
			strncpy(dataFile, argv[i], 1000);
		}
		else if (! strcmp(argv[i], "-ns"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing number of stages." << endl;
				return -1;
			}
			numberOfStages = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-nc"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing number of counters." << endl;
				return -1;
			}
			numberOfCounters = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-gs"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing grid size." << endl;
				return -1;
			}
			gridsize = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-sf"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing sampling factor." << endl;
				return -1;
			}
			sampling = true;
			samplingFactor = atof(argv[i]);
		}
		else if (! strcmp(argv[i], "-hz"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing horizon." << endl;
				return -1;
			}
			horizon = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-d"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing density." << endl;
				return -1;
			}
			density = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-qt"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing query type." << endl;
				return -1;
			}
			queryType = (strcasecmp(argv[i], "range") == 0) ? RANGE : SNAPSHOT;
		}
		else
		{
			cerr << "Usage: " << endl
				<< "  -df data file" << endl
				<< "  -gs gridsize" << endl
				<< "  -d density" << endl
				<< "  -ns number of stages" << endl
				<< "  -nc number of counters" << endl
				<< "  -sf sampling factor" << endl
				<< "  -hz horizon" << endl
				<< "  -qt query type (snapshot | range)" << endl;
			return -1;
		}
	}

	initRandom();

	double hashMod[numberOfStages];
	hashMod[0] = 0.6180339887;
	hashMod[1] = 0.5570895743;
	hashMod[2] = 0.4572980572;
	hashMod[3] = 0.3126508965;
	hashMod[4] = 0.2627098243;
	hashMod[5] = 0.1637089254;
	hashMod[6] = 0.0380575436;
	hashMod[7] = 0.7180974537;

	for (unsigned int cIndex = 8; cIndex < numberOfStages; cIndex++)
	{
		hashMod[cIndex] = nextUniformDouble(2.0 * std::numeric_limits<double>::epsilon(), (1.0 - 2.0 * std::numeric_limits<double>::epsilon()));
		cerr << "HashMod[" << cIndex << "]: " << hashMod[cIndex] << endl;
	}

	int T = TMIN;
	int t, ft, id, op, type, t1, t2, iun;
	float x, y, vx, vy, x1, x2, y1, y2;
	char file_buf[1000];
	map<int, MovingObject> dataset;
	map<int, MovingObject> sample;
	vector<Query> queries;
	double stepx = (XMAX - XMIN) / (double) gridsize;
	double stepy = (YMAX - YMIN) / (double) gridsize;
	map<int, Bucket> denseCells[horizon];
	unsigned long denseCellsSize = 0;
	unsigned long denseCellsSizeCount = 0;

	unsigned long*** estimator = 0;

	try
	{
		estimator = new unsigned long**[horizon];
		//unsigned long** estimator = new unsigned long*[numberOfStages];

		for (unsigned int cTime = 0; cTime < horizon; cTime++)
		{
			estimator[cTime] = new unsigned long*[numberOfStages];

			for (unsigned int cStage = 0; cStage < numberOfStages; cStage++)
			{
				estimator[cTime][cStage] = new unsigned long[numberOfCounters];
				//estimator[cStage] = new unsigned long[numberOfCounters];

				for (unsigned int cCounter = 0; cCounter < numberOfCounters; cCounter++)
				{
					estimator[cTime][cStage][cCounter] = 0;
					//estimator[cStage][cCounter] = 0;
				}
			}
		}
	}
	catch (...)
	{
		cerr << "Out of memory. Try running one more time." << endl;
		return -1;
	}

	//unsigned long maxCellID = ((gridsize - 1) * gridsize) + gridsize - 1;

	/*
	unsigned int* idGrid = new unsigned int[horizon * gridsize * gridsize];
	unsigned int cID = 0;
	for (unsigned int cTime = 0; cTime < horizon; cTime++)
	{
		for (unsigned int cX = 0; cX < gridsize; cX++)
		{
			for (unsigned int cY = 0; cY < gridsize; cY++)
			{
				idGrid[(cTime * gridsize * gridsize) + (cX * gridsize) + cY] = cID;
				cID++;
			}
		}
	}
	*/

	FILE* f = fopen(dataFile, "r");
	if (f == NULL)
	{
		cerr << "Cannot open " << dataFile << "." << endl;
		return -1;
	}

	cerr << "Data file: " << dataFile << ", gridsize: " << gridsize << ", density: " << density
		<< ", stepx: " << stepx << ", stepy: " << stepy << ", horizon: " << horizon
		<< ", stages: " << numberOfStages << ", counters: " << numberOfCounters << endl;

	time_t tstart = time(NULL);

	fgets(file_buf, 1000, f);

	int H = TMIN + horizon;

	while (! feof(f))
	{
		queries.clear();

		time_t ts = time(NULL);

		if (T >= H)
		{
			// rebuild the estimator. Copy paste here for now.
			H += horizon;

			for (unsigned int cTime = 0; cTime < horizon; cTime++)
			{
				denseCells[cTime].clear();

				for (unsigned int cStage = 0; cStage < numberOfStages; cStage++)
				{
					for (unsigned int cCounter = 0; cCounter < numberOfCounters; cCounter++)
					{
						estimator[cTime][cStage][cCounter] = 0;
					}
				}
			}

			map<int, MovingObject>* m;
			if (sampling) m = &sample;
			else m = &dataset;

			for (map<int, MovingObject>::iterator itData = m->begin(); itData != m->end(); itData++)
			{
				MovingObject& o = (*itData).second;

				for (unsigned int cTime = 0; cTime < horizon; cTime++)
				{
					int cX = (int) ((o.getX(H - horizon + cTime) - XMIN) / stepx);
					int cY = (int) ((o.getY(H - horizon + cTime) - YMIN) / stepy);

					if (cX >= 0 && cX < (int) gridsize && cY >= 0 && cY < (int) gridsize)
					{
						//unsigned long cellID = idGrid[(cTime * gridsize * gridsize) + (cX * gridsize) + cY];
						unsigned long cellID = (cX * gridsize) + cY;
						//double dCellID = ((double) cellID) / ((double) maxCellID);

						map<int, Bucket>::iterator itDenseCells = denseCells[cTime].find(cellID);
						if (itDenseCells != denseCells[cTime].end())
						{
							(*itDenseCells).second.m_count++;
						}
						else
						{
							unsigned int cDense = 0;

							int pCounters[numberOfStages];
							for (unsigned int cStage = 0; cStage < numberOfStages; cStage++)
							{
								int counter = h(cellID, numberOfCounters, hashMod[cStage]);
								//int counter = h2(dCellID, numberOfCounters, hashMod[cStage]);
								pCounters[cStage] = counter;
								//estimator[cTime][cStage][counter]++;
								if (estimator[cTime][cStage][counter] >= density - 1) cDense++;
							}

							if (cDense < numberOfStages)
							{
								for (unsigned int cStage = 0; cStage < numberOfStages; cStage++)
									estimator[cTime][cStage][pCounters[cStage]]++;

								/*
								// conservative updates
								// find the smallest counter.
								unsigned long cMin = estimator[0][pCounters[0]];
								int cMinIndex = 0;
								for (int cIndex = 1; cIndex < numberOfStages; cIndex++)
								{
									if (estimator[cIndex][pCounters[cIndex]] < cMin)
									{
										cMin = estimator[cIndex][pCounters[cIndex]];
										cMinIndex = cIndex;
									}
								}

								estimator[cMinIndex][pCounters[cMinIndex]]++;

								for (int cIndex = 0; cIndex < numberOfStages; cIndex++)
								{
									if (cIndex == cMinIndex) continue;
									estimator[cIndex][pCounters[cIndex]] = max(estimator[cIndex][pCounters[cIndex]], estimator[cMinIndex][pCounters[cMinIndex]]);
								}
								*/
							}
							else
							{
								Bucket b;
								b.m_low[0] = cX;
								b.m_high[0] = cX;
								b.m_low[1] = cY;
								b.m_high[1] = cY;
								b.m_count = density;
								denseCells[cTime].insert(pair<int, Bucket>(cellID, b));
							}
						}
					}
				}
			}
		}

		while (! feof(f))
		{
			sscanf(file_buf, "%d %d %d", &id, &op, &t);

			if (t > T) break;

			if (op == DELETE)
			{
				sscanf(file_buf, "%d %d %d %d %d %f %f %d %f %f", &id, &op, &t, &ft, &iun, &x, &vx, &iun, &y, &vy);
				MovingObject o(id, x, y, vx, vy, ft);
				dataset.erase(id);

				for (unsigned int cTime = T - (H - horizon); cTime < horizon; cTime++)
				{
					int cX = (int) ((o.getX(H - horizon + cTime) - XMIN) / stepx);
					int cY = (int) ((o.getY(H - horizon + cTime) - YMIN) / stepy);

					if (cX >= 0 && cX < (int) gridsize && cY >= 0 && cY < (int) gridsize)
					{
						//unsigned long cellID = idGrid[(cTime * gridsize * gridsize) + (cX * gridsize) + cY];
						unsigned long cellID = (cX * gridsize) + cY;
						//double dCellID = ((double) cellID) / ((double) maxCellID);

						map<int, Bucket>::iterator itDenseCells = denseCells[cTime].find(cellID);
						if (itDenseCells != denseCells[cTime].end())
						{
							if ((*itDenseCells).second.m_count == density)
							{
								denseCells[cTime].erase(itDenseCells);

								// this is wrong! cell ids that found the counters
								// above the threshold, never increased it by density.
								/*
								for (unsigned int cStage = 0; cStage < numberOfStages; cStage++)
								{
									int counter = h(cellID, numberOfCounters, hashMod[cStage]);
									//int counter = h2(dCellID, numberOfCounters, hashMod[cStage]);
									estimator[cTime][cStage][counter] -= density - 1;
								}
								*/
							}
							else
							{
								(*itDenseCells).second.m_count--;
							}
						}
						else
						{
							for (unsigned int cStage = 0; cStage < numberOfStages; cStage++)
							{
								int counter = h(cellID, numberOfCounters, hashMod[cStage]);
								//int counter = h2(dCellID, numberOfCounters, hashMod[cStage]);
								estimator[cTime][cStage][counter]--;

								/*
								// if a counter falls below the threshold, locate the according dense cells and remove them.
								if (estimator[cStage][counter] == density - 1)
								{
									for (int cIndex = 0; cIndex < horizon; cIndex++)
									{
										stack<int> del;
										for (map<int, Bucket>::iterator it = denseCells[cIndex].begin(); it != denseCells[cIndex].end(); it++)
										{
											int counter2 = h((*it).first, cStage, numberOfCounters);
											if (counter == counter2) del.push((*it).first);
										}

										while (! del.empty())
										{
											denseCells[cIndex].erase(del.top());
											del.pop();
										}
									}
								}
								*/
							}
						}
					}
				}
			}
			else if (op == QUERY)
			{
				// store queries for current time.
				sscanf(file_buf, "%d %d %d %d %f %f %f %f %d %d %d", &id, &op, &t, &ft, &x1, &x2, &y1, &y2, &t1, &t2, &type);
				Query q(x1, x2, y1, y2, t1, t2);
				queries.push_back(q);
			}
			else if (op == INSERT)
			{
				sscanf(file_buf, "%d %d %d %d %d %f %f %d %f %f", &id, &op, &t, &ft, &iun, &x, &vx, &iun, &y, &vy);
				MovingObject o(id, x, y, vx, vy, ft);
				dataset.erase(id);
				dataset.insert(pair<int, MovingObject>(id, o));

				if (sampling)
				{
					if (T == 0)
					{
						if (nextUniformDouble() < samplingFactor)
						{
							sample.erase(id);
							sample.insert(pair<int, MovingObject>(id, o));
						}
					}
					else
					{
						map<int, MovingObject>::iterator it = sample.find(o.m_id);
						if (it != sample.end())
						{
							sample.erase(id);
							sample.insert(pair<int, MovingObject>(id, o));
						}
					}
				}

				for (unsigned int cTime = T - (H - horizon); cTime < horizon; cTime++)
				{
					int cX = (int) ((o.getX(H - horizon + cTime) - XMIN) / stepx);
					int cY = (int) ((o.getY(H - horizon + cTime) - YMIN) / stepy);

					if (cX >= 0 && cX < (int) gridsize && cY >= 0 && cY < (int) gridsize)
					{
						//unsigned int cellID = idGrid[(cTime * gridsize * gridsize) + (cX * gridsize) + cY];
						unsigned int cellID = (cX * gridsize) + cY;
						//double dCellID = ((double) cellID) / ((double) maxCellID);

						map<int, Bucket>::iterator itDenseCells = denseCells[cTime].find(cellID);
						if (itDenseCells != denseCells[cTime].end())
						{
							(*itDenseCells).second.m_count++;
						}
						else
						{
							unsigned int cDense = 0;

							int pCounters[numberOfStages];
							for (unsigned int cStage = 0; cStage < numberOfStages; cStage++)
							{
								int counter = h(cellID, numberOfCounters, hashMod[cStage]);
								//int counter = h2(dCellID, numberOfCounters, hashMod[cStage]);
								pCounters[cStage] = counter;
								//estimator[cTime][cStage][counter]++;
								if (estimator[cTime][cStage][counter] >= density - 1) cDense++;
							}

							if (cDense < numberOfStages)
							{
								for (unsigned int cStage = 0; cStage < numberOfStages; cStage++)
									estimator[cTime][cStage][pCounters[cStage]]++;

								// conservative updates.
								/*
								// find the smallest counter.
								unsigned long cMin = estimator[0][pCounters[0]];
								int cMinIndex = 0;
								for (int cIndex = 1; cIndex < numberOfStages; cIndex++)
								{
									if (estimator[cIndex][pCounters[cIndex]] < cMin)
									{
										cMin = estimator[cIndex][pCounters[cIndex]];
										cMinIndex = cIndex;
									}
								}

								estimator[cMinIndex][pCounters[cMinIndex]]++;

								for (int cIndex = 0; cIndex < numberOfStages; cIndex++)
								{
									if (cIndex == cMinIndex) continue;
									estimator[cIndex][pCounters[cIndex]] = max(estimator[cIndex][pCounters[cIndex]], estimator[cMinIndex][pCounters[cMinIndex]]);
								}
								*/
							}
							else
							{
								Bucket b;
								b.m_low[0] = cX;
								b.m_high[0] = cX;
								b.m_low[1] = cY;
								b.m_high[1] = cY;
								b.m_count = density;
								denseCells[cTime].insert(pair<int, Bucket>(cellID, b));
							}
						}
					}
				}
			}

			fgets(file_buf, 1000, f);
		}

		time_t te = time(NULL);
		updateTime += te - ts;

		cerr << "Time: " << T << ", Queries: " << queries.size() << ", dataset size: " << dataset.size() << endl;

		for (unsigned int cTime = 0; cTime < horizon; cTime++)
		{
			//cerr << "Dense cells for time " << cTime << ": " << denseCells[cTime].size() << endl;
			denseCellsSize += denseCells[cTime].size();
		}
		denseCellsSizeCount++;

		ts = time(NULL);

		if (queryType == SNAPSHOT)
		{
			for (unsigned int cQuery = 0; cQuery < queries.size(); cQuery++)
			{
				if (queries[cQuery].m_t1 >= H)
				{
					cout << "ERROR. Time: " << T << ", Query: " << cQuery << " out of estimator horizon. " << queries[cQuery].m_t1 << endl;
					continue;
				}

				int cTime = queries[cQuery].m_t1 - (H - horizon);

				for (map<int, Bucket>::iterator it = denseCells[cTime].begin(); it != denseCells[cTime].end(); it++)
				{
					cout << T << " " << cQuery << " " << (*it).second.m_low[0] << " " << (*it).second.m_low[1] << " " << (*it).second.m_count << " " << queries[cQuery].m_t1 << endl;
				}
			}
		}
		else
		{
			set<int> found[horizon];
			for (unsigned int cTime = T - (H - horizon); cTime < horizon; cTime++)
			{
				for (map<int, Bucket>::iterator itDenseCells = denseCells[cTime].begin(); itDenseCells != denseCells[cTime].end(); itDenseCells++)
				{
					int cellID = (int) (*itDenseCells).second.m_low[0] * gridsize + (int) (*itDenseCells).second.m_low[1];
					set<int>::iterator itFound = found[cTime].find(cellID);
					if (itFound != found[cTime].end()) continue;

					int interval = 1;
					for (unsigned int cNext = cTime + 1; cNext < horizon; cNext++)
					{
						map<int, Bucket>::iterator itDenseCells2 = denseCells[cNext].find(cellID);
						if (itDenseCells2 != denseCells[cNext].end()) interval++;
						else break;
					}
					found[cTime].insert(cellID);
					cout << T << " " << (*itDenseCells).second.m_low[0] << " " << (*itDenseCells).second.m_low[1] << " "
						<< (*itDenseCells).second.m_count << " " << interval << endl;
				}
			}
		}

		te = time(NULL);
		queryTime += te - ts;

		T++;
	}

	time_t tend = time(NULL);
	time_t totalTime = tend - tstart;

	// 2 bytes is enough for the counters for up to density D=16000.
	// I report the average estimator size here.
	double dSize = ((numberOfStages * numberOfCounters * horizon) / 2.0) +  ((double) denseCellsSize / (double) denseCellsSizeCount);

	cerr << "Total Time: " << totalTime << endl;
	cerr << "Query Time: " << queryTime << endl;
	cerr << "Update Time: " << updateTime << endl;
	cerr << "Estimator size: " << dSize << endl;

	cout << "Info: " << dataset.size() << " " << gridsize << " " << density << " " << horizon << " " << totalTime << " " << queryTime << " "
		<< updateTime << " " <<  dSize << endl;

	for (unsigned int cTime = 0; cTime < horizon; cTime++)
	{
		for (unsigned int cStage = 0; cStage < numberOfStages; cStage++)
		{
			delete[] estimator[cTime][cStage];
		}
		delete[] estimator[cTime];
	}
	delete[] estimator;

	//delete[] idGrid;

	return 0;
}
