// On-Line Discovery of Dense Areas in Spatio-temporal Databases
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

/*
	Find all future cell densities after a specific time instant, during a specified horizon.
*/

#include "global.h"

class Entry
{
	public:
		int x, y, f, qt;
		Entry(int a, int b, int c, int d) : x(a), y(b), f(c), qt(d) {}
};

int main(int argc, char** argv)
{
	char dataFile[1000] = "d";
	int gridsize = 250;
	unsigned int horizon = 50;
	int snapTime = 0;

	for (int i = 1; i < argc; i++)
	{
		if (! strcmp(argv[i], "-df"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing data file." << endl;
				return -1;
			}
			strncpy(dataFile, argv[i], 1000);
		}
		else if (! strcmp(argv[i], "-st"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing snapshot time." << endl;
				return -1;
			}
			snapTime = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-gs"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing grid size." << endl;
				return -1;
			}
			gridsize = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-hz"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing horizon." << endl;
				return -1;
			}
			horizon = atoi(argv[i]);
		}
		else
		{
			cerr << "Usage: " << endl
				<< "  -df data file" << endl
				<< "  -st snapshot time" << endl
				<< "  -gs gridsize" << endl
				<< "  -hz horizon" << endl;
			return -1;
		}
	}

	initRandom();

	int T = 0;
	int t, ft, id, op, iun;
	float x, y, vx, vy;
	char file_buf[1000];
	map<int, MovingObject> dataset;
	long* grid = new long[horizon * gridsize * gridsize];
	double stepx = (XMAX - XMIN) / gridsize;
	double stepy = (YMAX - YMIN) / gridsize;

	FILE* f = fopen(dataFile, "r");
	if (f == 0)
	{
		cerr << "Cannot open " << dataFile << "." << endl;
		return -1;
	}

	for (unsigned int cTime = 0; cTime < horizon; cTime++)
	{
		for (int cX = 0; cX < gridsize; cX++)
		{
			for (int cY = 0; cY < gridsize; cY++)
			{
				grid[(cTime * gridsize * gridsize) + (cX * gridsize) + cY] = 0;
			}
		}
	}

	cerr << "Data file: " << dataFile << ", gridsize: " << gridsize << ", horizon: " << horizon
		<< ", stepx: " << stepx << ", stepy: " << stepy << endl;

	fgets(file_buf, 1000, f);

	int H = TMIN + horizon;

	while (! feof(f))
	{
		if (T >= H)
		{
			// rebuild the estimator. Copy paste here for now.
			H += horizon;

			for (unsigned int cTime = 0; cTime < horizon; cTime++)
			{
				for (int cX = 0; cX < gridsize; cX++)
				{
					for (int cY = 0; cY < gridsize; cY++)
					{
						grid[(cTime * gridsize * gridsize) + (cX * gridsize) + cY] = 0;
					}
				}
			}

			for (map<int, MovingObject>::iterator it = dataset.begin(); it != dataset.end(); it++)
			{
				MovingObject& o = (*it).second;

				for (unsigned int cTime = 0; cTime < horizon; cTime++)
				{
					double ox = o.getX(H - horizon + cTime);
					double oy = o.getY(H - horizon + cTime);

					int cX = (int) (ox / stepx);
					int cY = (int) (oy / stepy);

					if (cX >= 0 && cX < gridsize && cY >= 0 && cY < gridsize)
					{
						grid[(cTime * gridsize * gridsize) + (cX * gridsize) + cY]++;
					}
				}
			}
		}

		while (! feof(f))
		{
			sscanf(file_buf, "%d %d %d", &id, &op, &t);

			if (t > T) break;

			if (op == DELETE)
			{
				sscanf(file_buf, "%d %d %d %d %d %f %f %d %f %f", &id, &op, &t, &ft, &iun, &x, &vx, &iun, &y, &vy);
				MovingObject o(id, x, y, vx, vy, ft);
				dataset.erase(id);

				for (unsigned int cTime = T - (H - horizon); cTime < horizon; cTime++)
				{
					double ox = o.getX(H - horizon + cTime);
					double oy = o.getY(H - horizon + cTime);

					int cX = (int) (ox / stepx);
					int cY = (int) (oy / stepy);

					if (cX >= 0 && cX < gridsize && cY >= 0 && cY < gridsize)
					{
						grid[(cTime * gridsize * gridsize) + (cX * gridsize) + cY]--;
					}
				}
			}
			else if (op == QUERY)
			{
			}
			else if (op == INSERT)
			{
				sscanf(file_buf, "%d %d %d %d %d %f %f %d %f %f", &id, &op, &t, &ft, &iun, &x, &vx, &iun, &y, &vy);
				MovingObject o(id, x, y, vx, vy, ft);
				dataset.erase(id);
				dataset.insert(pair<int, MovingObject>(id, o));

				for (unsigned int cTime = T - (H - horizon); cTime < horizon; cTime++)
				{
					double ox = o.getX(H - horizon + cTime);
					double oy = o.getY(H - horizon + cTime);

					int cX = (int) (ox / stepx);
					int cY = (int) (oy / stepy);

					if (cX >= 0 && cX < gridsize && cY >= 0 && cY < gridsize)
					{
						grid[(cTime * gridsize * gridsize) + (cX * gridsize) + cY]++;
					}
				}
			}

			fgets(file_buf, 1000, f);
		}

		cerr << "Time: " << T << ", dataset size: " << dataset.size() << endl;

		if (T == snapTime)
		{
			unsigned long densities[gridsize][gridsize];
			for (int cX = 0; cX < gridsize; cX++)
				for (int cY = 0; cY < gridsize; cY++)
					densities[cX][cY] = 0;

			for (int cTime = T - (H - horizon); cTime < horizon; cTime++)
			{
				for (int cX = 0; cX < gridsize; cX++)
				{
					for (int cY = 0; cY < gridsize; cY++)
					{
						densities[cX][cY] += grid[cTime * gridsize * gridsize + cX * gridsize + cY];
					}
				}
			}

			for (int cX = 0; cX < gridsize; cX++)
				for (int cY = 0; cY < gridsize; cY++)
					cout << cX << " " << cY << " " << densities[cX][cY] << endl;

			break;
		}

		T++;
	}

	delete[] grid;
	return 0;
}
