// On-Line Discovery of Dense Areas in Spatio-temporal Databases
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

#include "global.h"

int main(int argc, char** argv)
{
	char dataFile[1000] = "d";
	int gridsize = 250;
	unsigned int density = 1000;
	unsigned int horizon = 20;
	unsigned int gridGranularity = 4;
	int queryType = SNAPSHOT;

	time_t queryTime = 0;
	time_t updateTime = 0;

	for (int i = 1; i < argc; i++)
	{
		if (! strcmp(argv[i], "-df"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing data file." << endl;
				return -1;
			}
			strncpy(dataFile, argv[i], 1000);
		}
		else if (! strcmp(argv[i], "-gs"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing grid size." << endl;
				return -1;
			}
			gridsize = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-d"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing density." << endl;
				return -1;
			}
			density = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-hz"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing horizon." << endl;
				return -1;
			}
			horizon = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-gr"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing grid granularity." << endl;
				return -1;
			}
			gridGranularity = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-qt"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing query type." << endl;
				return -1;
			}
			queryType = (strcasecmp(argv[i], "range") == 0) ? RANGE : SNAPSHOT;
		}
		else
		{
			cerr << "Usage: " << endl
				<< "  -df data file" << endl
				<< "  -gs gridsize" << endl
				<< "  -hz horizon" << endl
				<< "  -gr grid granularity" << endl
				<< "  -d density" << endl
				<< "  -qt query type (snapshot | range)" << endl;
			return -1;
		}
	}

	if (horizon % gridGranularity != 0)
	{
		cerr << "Granularity should be a divisor of horizon." << endl;
		return -1;
	}

	initRandom();

	int T = TMIN;
	unsigned int granules = horizon / gridGranularity;
	int t, ft, id, op, type, t1, t2, iun;
	float x, y, vx, vy, x1, x2, y1, y2;
	char file_buf[1000];
	map<int, MovingObject> dataset;
	vector<Query> queries;
	map<int, Bucket> estimator[granules];
	double stepx = (XMAX - XMIN) / gridsize;
	double stepy = (YMAX - YMIN) / gridsize;
	unsigned long averageEstimatorSize = 0;
	unsigned long averageEstimatorSizeCount = 0;

	FILE* f = fopen(dataFile, "r");
	if (f == 0)
	{
		cerr << "Cannot open " << dataFile << "." << endl;
		return -1;
	}

	cerr << "Data file: " << dataFile << ", gridsize: " << gridsize << ", horizon: " << horizon << ", density: " << density
		<< ", stepx: " << stepx << ", stepy: " << stepy << ", granularity: " << gridGranularity << endl;

	time_t tstart = time(NULL);

	fgets(file_buf, 1000, f);

	int H = TMIN + horizon;

	while (! feof(f))
	{
		queries.clear();

		time_t ts = time(NULL);

		if (T >= H)
		{
			// rebuild the estimator. Copy paste here for now.
			H += horizon;

			for (unsigned int cGranule = 0; cGranule < granules; cGranule++)
			{
				estimator[cGranule].clear();
			}

			for (map<int, MovingObject>::iterator it = dataset.begin(); it != dataset.end(); it++)
			{
				MovingObject& o = (*it).second;

				for (unsigned int cTime = 0; cTime < horizon; cTime += gridGranularity)
				{
					int cXPrev = -1;
					int cYPrev = -1;

					int estimatorIndex = (int) floor((double) cTime / (double) gridGranularity);

					for (unsigned int cGranule = 0; cGranule < gridGranularity; cGranule++)
					{
						int cX = (int) ((o.getX(H - horizon + cTime + cGranule) - XMIN) / stepx);
						int cY = (int) ((o.getY(H - horizon + cTime + cGranule) - YMIN) / stepy);

						if (cX >= 0 && cX < gridsize && cY >= 0 && cY < gridsize)
						{
							if (cX == cXPrev && cY == cYPrev) continue;

							int cellID = (cX * gridsize) + cY;

							map<int, Bucket>::iterator it = estimator[estimatorIndex].find(cellID);
							if (it != estimator[estimatorIndex].end())
							{
								(*it).second.m_count++;
							}
							else
							{
								Bucket b;
								b.m_low[0] = cX;
								b.m_high[0] = cX;
								b.m_low[1] = cY;
								b.m_high[1] = cY;
								b.m_count = 1;
								estimator[estimatorIndex].insert(pair<int, Bucket>(cellID, b));
							}
						}

						cXPrev = cX;
						cYPrev = cY;
					}
				}
			}
		}

		while (! feof(f))
		{
			sscanf(file_buf, "%d %d %d", &id, &op, &t);

			if (t > T) break;

			if (op == DELETE)
			{
				sscanf(file_buf, "%d %d %d %d %d %f %f %d %f %f", &id, &op, &t, &ft, &iun, &x, &vx, &iun, &y, &vy);
				MovingObject o(id, x, y, vx, vy, ft);
				dataset.erase(id);

				unsigned int snapTime = (unsigned int) floor((double) (T - (H - horizon)) / (double) gridGranularity) * gridGranularity;
				unsigned int timePassed = (T - (H - horizon)) - snapTime;

				for (unsigned int cTime = snapTime; cTime < horizon; cTime += gridGranularity)
				{
					int cXPrev = -1;
					int cYPrev = -1;

					int estimatorIndex = (int) floor((double) cTime / (double) gridGranularity);

					for (unsigned int cGranule = timePassed; cGranule < gridGranularity; cGranule++)
					{
						int cX = (int) ((o.getX(H - horizon + cTime + cGranule) - XMIN) / stepx);
						int cY = (int) ((o.getY(H - horizon + cTime + cGranule) - YMIN) / stepy);

						timePassed = 0;

						if (cX >= 0 && cX < gridsize && cY >= 0 && cY < gridsize)
						{
							if (cX == cXPrev && cY == cYPrev) continue;

							int cellID = (cX * gridsize) + cY;

							map<int, Bucket>::iterator it = estimator[estimatorIndex].find(cellID);
							if (it != estimator[estimatorIndex].end())
							{
								(*it).second.m_count--;
							}
						}

						cXPrev = cX;
						cYPrev = cY;
					}
				}
			}
			else if (op == QUERY)
			{
				// store queries for current time.
				sscanf(file_buf, "%d %d %d %d %f %f %f %f %d %d %d", &id, &op, &t, &ft, &x1, &x2, &y1, &y2, &t1, &t2, &type);
				Query q(x1, x2, y1, y2, t1, t2);
				queries.push_back(q);
			}
			else if (op == INSERT)
			{
				sscanf(file_buf, "%d %d %d %d %d %f %f %d %f %f", &id, &op, &t, &ft, &iun, &x, &vx, &iun, &y, &vy);
				MovingObject o(id, x, y, vx, vy, ft);
				dataset.erase(id);
				dataset.insert(pair<int, MovingObject>(id, o));

				unsigned int snapTime = (unsigned int) floor((double) (T - (H - horizon)) / (double) gridGranularity) * gridGranularity;
				unsigned int timePassed = (T - (H - horizon)) - snapTime;

				for (unsigned int cTime = snapTime; cTime < horizon; cTime += gridGranularity)
				{
					int cXPrev = -1;
					int cYPrev = -1;

					int estimatorIndex = (int) floor((double) cTime / (double) gridGranularity);

					for (unsigned int cGranule = timePassed; cGranule < gridGranularity; cGranule++)
					{
						int cX = (int) ((o.getX(H - horizon + cTime + cGranule) - XMIN) / stepx);
						int cY = (int) ((o.getY(H - horizon + cTime + cGranule) - YMIN) / stepy);

						timePassed = 0;

						if (cX >= 0 && cX < gridsize && cY >= 0 && cY < gridsize)
						{
							if (cX == cXPrev && cY == cYPrev) continue;

							int cellID = (cX * gridsize) + cY;

							map<int, Bucket>::iterator it = estimator[estimatorIndex].find(cellID);
							if (it != estimator[estimatorIndex].end())
							{
								(*it).second.m_count++;
							}
							else
							{
								Bucket b;
								b.m_low[0] = cX;
								b.m_high[0] = cX;
								b.m_low[1] = cY;
								b.m_high[1] = cY;
								b.m_count = 1;
								estimator[estimatorIndex].insert(pair<int, Bucket>(cellID, b));
							}
						}

						cXPrev = cX;
						cYPrev = cY;
					}
				}
			}

			fgets(file_buf, 1000, f);
		}

		time_t te = time(NULL);
		updateTime += te - ts;

		cerr << "Time: " << T << ", Queries: " << queries.size() << ", dataset size: " << dataset.size() << endl;

		for (unsigned int cGranule = 0; cGranule < granules; cGranule++)
		{
			averageEstimatorSize  += estimator[cGranule].size();
		}
		averageEstimatorSizeCount++;

		ts= time(NULL);

		if (queryType == SNAPSHOT)
		{
			for (unsigned int cQuery = 0; cQuery < queries.size(); cQuery++)
			{
				if (queries[cQuery].m_t1 >= H)
				{
					cout << "ERROR. Time: " << T << ", Query: " << cQuery << " out of estimator horizon. " << queries[cQuery].m_t1 << endl;
					continue;
				}

				int estimatorIndex = (int) floor((double) (queries[cQuery].m_t1 - (H - horizon)) / (double) gridGranularity);

				for (map<int, Bucket>::iterator it = estimator[estimatorIndex].begin(); it != estimator[estimatorIndex].end(); it++)
				{
					if ((*it).second.m_count >= density)
					{
						cout << T << " " << cQuery << " " << (*it).second.m_low[0] << " " << (*it).second.m_low[1] << " " << (*it).second.m_count << " " << queries[cQuery].m_t1 << endl;
					}
				}
			}
		}
		else
		{
			set<int> found[horizon];
			for (unsigned int cTime = T - (H - horizon); cTime < horizon; cTime++)
			{
				int estimatorIndex = (int) floor((double) (cTime) / (double) gridGranularity);

				for (map<int, Bucket>::iterator it = estimator[estimatorIndex].begin(); it != estimator[estimatorIndex].end(); it++)
				{
					int cellID = (int) (*it).second.m_low[0] * gridsize + (int) (*it).second.m_low[1];
					if ((*it).second.m_count < density) continue;

					set<int>::iterator itFound = found[cTime].find(cellID);
					if (itFound != found[cTime].end()) continue;

					int interval = 1;
					for (unsigned int cNext = estimatorIndex + 1; cNext < granules; cNext++)
					{
						map<int, Bucket>::iterator it2 = estimator[cNext].find(cellID);
						if (it2 != estimator[cNext].end() && (*it2).second.m_count >= density) interval += gridGranularity;
						else break;
					}
					found[cTime].insert(cellID);
					cout << T << " " << (*it).second.m_low[0] << " " << (*it).second.m_low[1] << " " << (*it).second.m_count << " " << interval << endl;
				}
			}
		}

		te = time(NULL);
		queryTime += te - ts;

		T++;
	}

	time_t tend = time(NULL);
	time_t totalTime = tend - tstart;

	cerr << "Total Time: " << totalTime << endl;
	cerr << "Query Time: " << queryTime << endl;
	cerr << "Update Time: " << updateTime << endl;

	cout << "Info: " << dataset.size() << " " << gridsize << " " << density << " " << horizon << " " << totalTime << " " << queryTime << " "
		<< updateTime << " " << (double) averageEstimatorSize / (double) averageEstimatorSizeCount << endl;
}
