// On-Line Discovery of Dense Areas in Spatio-temporal Databases
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

/*
	Prints the dense cells at time snapTime in gnuplot format.
	Extrapolates the trajectories of all objects at snapTime, given
	the latest update information.
*/

#include "global.h"

int main(int argc, char** argv)
{
	char dataFile[1000] = "d";
	unsigned int gridsize = 250;
	unsigned int density = 1000;
	int snapTime = 0;

	for (int i = 1; i < argc; i++)
	{
		if (! strcmp(argv[i], "-df"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing data file." << endl;
				return -1;
			}
			strncpy(dataFile, argv[i], 1000);
		}
		else if (! strcmp(argv[i], "-st"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing snapshot time." << endl;
				return -1;
			}
			snapTime = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-gs"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing grid size." << endl;
				return -1;
			}
			gridsize = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-d"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing density." << endl;
				return -1;
			}
			density = atoi(argv[i]);
		}
		else
		{
			cerr << "Usage: " << endl
				<< "  -df data file" << endl
				<< "  -st snapshot time" << endl
				<< "  -gs gridsize" << endl
				<< "  -d density" << endl;
			return -1;
		}
	}

	initRandom();

	int T = 0;
	int t, ft, id, op, iun;
	float x, y, vx, vy;
	char file_buf[1000];
	map<int, MovingObject> dataset;
	Bucket* grid = new Bucket[gridsize * gridsize];
	double stepx = (XMAX - XMIN) / gridsize;
	double stepy = (YMAX - YMIN) / gridsize;

	FILE* f = fopen(dataFile, "r");
	if (f == NULL)
	{
		cerr << "Cannot open " << dataFile << "." << endl;
		return -1;
	}

	for (unsigned int cX = 0; cX < gridsize; cX++)
	{
		for (unsigned int cY = 0; cY < gridsize; cY++)
		{
			grid[cX * gridsize + cY].m_count = 0;
			grid[cX * gridsize + cY].m_low[0] = XMIN + ((double) cX * stepx);
			grid[cX * gridsize + cY].m_high[0] = XMIN + ((double) cX * stepx) + stepx;
			grid[cX * gridsize + cY].m_low[1] = YMIN + ((double) cY * stepy);
			grid[cX * gridsize + cY].m_high[1] = YMIN + ((double) cY * stepy) + stepy;
		}
	}

	cerr << "Data file: " << dataFile << ", gridsize: " << gridsize << ", density: " << density
		<< ", stepx: " << stepx << ", stepy: " << stepy << endl;

	fgets(file_buf, 1000, f);

	while (! feof(f))
	{
		while (! feof(f))
		{
			sscanf(file_buf, "%d %d %d", &id, &op, &t);

			if (t > T) break;

			if (op == DELETE)
			{
			}
			else if (op == QUERY)
			{
			}
			else if (op == INSERT)
			{
				sscanf(file_buf, "%d %d %d %d %d %f %f %d %f %f", &id, &op, &t, &ft, &iun, &x, &vx, &iun, &y, &vy);
				MovingObject o(id, x, y, vx, vy, ft);
				dataset.erase(id);
				dataset.insert(pair<int, MovingObject>(id, o));
			}

			fgets(file_buf, 1000, f);
		}

		cerr << "Time: " << T << ", dataset size: " << dataset.size() << endl;

		if (T == snapTime)
		{
			for (map<int, MovingObject>::iterator it = dataset.begin(); it != dataset.end(); it++)
			{
				MovingObject& o = (*it).second;

				unsigned int cX = (unsigned int) ((o.getX(T) - XMIN) / stepx);
				unsigned int cY = (unsigned int) ((o.getY(T) - YMIN) / stepy);

				// not all extrapolations are inside the grid at query time.
				if (cX >= 0 && cX < gridsize && cY >= 0 && cY < gridsize) grid[cX * gridsize + cY].m_count++;
			}

			for (unsigned int cX = 0; cX < gridsize; cX++)
			{
				for (unsigned int cY = 0; cY < gridsize; cY++)
				{
					if (grid[cX * gridsize + cY].m_count >= density)
					{
						cerr << cX << ":" << cY << "=" << grid[cX * gridsize + cY].m_count << endl;
						cout << grid[cX * gridsize + cY].m_low[0] << " " << grid[cX * gridsize + cY].m_low[1] << endl;
						cout << grid[cX * gridsize + cY].m_high[0] << " " << grid[cX * gridsize + cY].m_low[1] << endl;
						cout << grid[cX * gridsize + cY].m_high[0] << " " << grid[cX * gridsize + cY].m_high[1] << endl;
						cout << grid[cX * gridsize + cY].m_low[0] << " " << grid[cX * gridsize + cY].m_high[1] << endl;
						cout << grid[cX * gridsize + cY].m_low[0] << " " << grid[cX * gridsize + cY].m_low[1] << endl << endl;
					}
				}
			}

			delete[] grid;
			return 0;
		}

		T++;
	}

	delete[] grid;
}
