// On-Line Discovery of Dense Areas in Spatio-temporal Databases
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

#include "global.h"

struct maxDensity : public binary_function<Bucket, Bucket, bool>
{
	bool operator()(Bucket __x, Bucket __y) const { return __x.m_count < __y.m_count; }
};

int main(int argc, char** argv)
{
	char dataFile[1000] = "d";
	int gridsize = 250;
	unsigned int density = 1000;
	unsigned int horizon = 20;
	int queryType = SNAPSHOT;

	time_t queryTime = 0;
	time_t updateTime = 0;

	for (int i = 1; i < argc; i++)
	{
		if (! strcmp(argv[i], "-df"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing data file." << endl;
				return -1;
			}
			strncpy(dataFile, argv[i], 1000);
		}
		else if (! strcmp(argv[i], "-gs"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing grid size." << endl;
				return -1;
			}
			gridsize = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-d"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing density." << endl;
				return -1;
			}
			density = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-hz"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing horizon." << endl;
				return -1;
			}
			horizon = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-qt"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing query type." << endl;
				return -1;
			}
			if (strcasecmp(argv[i], "range") == 0) queryType = RANGE;
			else if (strcasecmp(argv[i], "topk") == 0) queryType = TOPK;
			else queryType = SNAPSHOT;
		}
		else
		{
			cerr << "Usage: " << endl
				<< "  -df data file" << endl
				<< "  -gs gridsize" << endl
				<< "  -hz horizon" << endl
				<< "  -d density" << endl
				<< "  -qt query type (snapshot | range)" << endl;
			return -1;
		}
	}

	initRandom();

	int T = TMIN;
	int t, ft, id, op, type, t1, t2, iun;
	float x, y, vx, vy, x1, x2, y1, y2;
	char file_buf[1000];
	map<int, MovingObject> dataset;
	vector<Query> queries;
	unsigned int* grid = new unsigned int[horizon * gridsize * gridsize];
	double stepx = (XMAX - XMIN) / gridsize;
	double stepy = (YMAX - YMIN) / gridsize;

	FILE* f = fopen(dataFile, "r");
	if (f == 0)
	{
		cerr << "Cannot open " << dataFile << "." << endl;
		return -1;
	}

	for (unsigned int cTime = 0; cTime < horizon; cTime++)
	{
		for (int cX = 0; cX < gridsize; cX++)
		{
			for (int cY = 0; cY < gridsize; cY++)
			{
				grid[(cTime * gridsize * gridsize) + (cX * gridsize) + cY] = 0;
			}
		}
	}

	cerr << "Data file: " << dataFile << ", gridsize: " << gridsize << ", horizon: " << horizon << ", density: " << density
		<< ", stepx: " << stepx << ", stepy: " << stepy << ", querytype: " << queryType << endl;

	time_t tstart = time(NULL);

	fgets(file_buf, 1000, f);

	int H = TMIN + horizon;

	while (! feof(f))
	{
		queries.clear();

		time_t ts = time(NULL);

		if (T >= H)
		{
			// rebuild the estimator. Copy paste here for now.
			H += horizon;

			for (unsigned int cTime = 0; cTime < horizon; cTime++)
			{
				for (int cX = 0; cX < gridsize; cX++)
				{
					for (int cY = 0; cY < gridsize; cY++)
					{
						grid[(cTime * gridsize * gridsize) + (cX * gridsize) + cY] = 0;
					}
				}
			}

			for (map<int, MovingObject>::iterator it = dataset.begin(); it != dataset.end(); it++)
			{
				MovingObject& o = (*it).second;

				for (unsigned int cTime = 0; cTime < horizon; cTime++)
				{
					int cX = (int) ((o.getX(H - horizon + cTime) - XMIN) / stepx);
					int cY = (int) ((o.getY(H - horizon + cTime) - YMIN) / stepy);

					if (cX >= 0 && cX < gridsize && cY >= 0 && cY < gridsize)
					{
						grid[(cTime * gridsize * gridsize) + (cX * gridsize) + cY]++;
					}
				}
			}
		}

		while (! feof(f))
		{
			sscanf(file_buf, "%d %d %d", &id, &op, &t);

			if (t > T) break;

			if (op == DELETE)
			{
				sscanf(file_buf, "%d %d %d %d %d %f %f %d %f %f", &id, &op, &t, &ft, &iun, &x, &vx, &iun, &y, &vy);
				MovingObject o(id, x, y, vx, vy, ft);
				dataset.erase(id);

				for (unsigned int cTime = T - (H - horizon); cTime < horizon; cTime++)
				{
					int cX = (int) ((o.getX(H - horizon + cTime) - XMIN) / stepx);
					int cY = (int) ((o.getY(H - horizon + cTime) - YMIN) / stepy);

					if (cX >= 0 && cX < gridsize && cY >= 0 && cY < gridsize)
					{
						grid[(cTime * gridsize * gridsize) + (cX * gridsize) + cY]--;
					}
				}
			}
			else if (op == QUERY)
			{
				// store queries for current time.
				sscanf(file_buf, "%d %d %d %d %f %f %f %f %d %d %d", &id, &op, &t, &ft, &x1, &x2, &y1, &y2, &t1, &t2, &type);
				Query q(x1, x2, y1, y2, t1, t2);
				queries.push_back(q);
			}
			else if (op == INSERT)
			{
				sscanf(file_buf, "%d %d %d %d %d %f %f %d %f %f", &id, &op, &t, &ft, &iun, &x, &vx, &iun, &y, &vy);
				MovingObject o(id, x, y, vx, vy, ft);
				dataset.erase(id);
				dataset.insert(pair<int, MovingObject>(id, o));

				for (unsigned int cTime = T - (H - horizon); cTime < horizon; cTime++)
				{
					int cX = (int) ((o.getX(H - horizon + cTime) - XMIN) / stepx);
					int cY = (int) ((o.getY(H - horizon + cTime) - YMIN) / stepy);

					if (cX >= 0 && cX < gridsize && cY >= 0 && cY < gridsize)
					{
						grid[(cTime * gridsize * gridsize) + (cX * gridsize) + cY]++;
					}
				}
			}

			fgets(file_buf, 1000, f);
		}

		time_t te = time(NULL);
		updateTime += te - ts;

		cerr << "Time: " << T << ", Queries: " << queries.size() << ", dataset size: " << dataset.size() << endl;

		ts= time(NULL);

		if (queryType == SNAPSHOT)
		{
			for (unsigned int cQuery = 0; cQuery < queries.size(); cQuery++)
			{
				if (queries[cQuery].m_t1 >= H)
				{
					cout << "ERROR. Time: " << T << ", Query: " << cQuery << " out of estimator horizon. " << queries[cQuery].m_t1 << endl;
					continue;
				}

				for (int cX = 0; cX < gridsize; cX++)
				{
					for (int cY = 0; cY < gridsize; cY++)
					{
						if (grid[((queries[cQuery].m_t1 - (H - horizon)) * gridsize * gridsize) + (cX * gridsize) + cY] >= density)
						{
							cout << T << " " << cQuery << " " << cX << " " << cY << " " << grid[((queries[cQuery].m_t1 - (H - horizon)) * gridsize * gridsize) + (cX * gridsize) + cY] << endl;

							//cout << grid[cX * gridsize + cY].low[0] << " " << grid[cX * gridsize + cY].low[1] << endl;
							//cout << grid[cX * gridsize + cY].high[0] << " " << grid[cX * gridsize + cY].low[1] << endl;
							//cout << grid[cX * gridsize + cY].high[0] << " " << grid[cX * gridsize + cY].high[1] << endl;
							//cout << grid[cX * gridsize + cY].low[0] << " " << grid[cX * gridsize + cY].high[1] << endl;
							//cout << grid[cX * gridsize + cY].low[0] << " " << grid[cX * gridsize + cY].low[1] << endl << endl;
						}
					}
				}
			}
		}
		else if (queryType == RANGE)
		{
			set<int> found[horizon];
			for (unsigned int cTime = T - (H - horizon); cTime < horizon; cTime++)
			{
				for (int cX = 0; cX < gridsize; cX++)
				{
					for (int cY = 0; cY < gridsize; cY++)
					{
						int cellID = (cTime * gridsize * gridsize) + (cX * gridsize) + cY;
						if (grid[cellID] < density) continue;

						set<int>::iterator itFound = found[cTime].find(cellID);
						if (itFound != found[cTime].end()) continue;

						int interval = 1;
						for (unsigned int cNext = cTime + 1; cNext < horizon; cNext++)
						{
							if (grid[(cNext * gridsize * gridsize) + (cX * gridsize) + cY] >= density)
							{
								interval++;
								found[cNext].insert(cellID);
							}
							else break;
						}
						found[cTime].insert(cellID);
						cout << T << " " << cX << " " << cY << " " << grid[cellID] << " " << interval << endl;
					}
				}
			}
		}
		else
		{
			for (unsigned int cQuery = 0; cQuery < queries.size(); cQuery++)
			{
				if (queries[cQuery].m_t1 >= H)
				{
					cout << "ERROR. Time: " << T << ", Query: " << cQuery << " out of estimator horizon. " << queries[cQuery].m_t1 << endl;
					continue;
				}

				priority_queue<Bucket, vector<Bucket>, maxDensity> pq;

				for (unsigned int cTime = queries[cQuery].m_t1 - (H - horizon); cTime < horizon; cTime++)
				{
					for (int cX = 0; cX < gridsize; cX++)
					{
						for (int cY = 0; cY < gridsize; cY++)
						{
							Bucket b;
							b.m_low[0] = cX;
							b.m_high[0] = cX;
							b.m_low[1] = cY;
							b.m_high[1] = cY;
							b.m_count = grid[(cTime * gridsize * gridsize) + (cX * gridsize) + cY];

							pq.push(b);
						}
					}
				}

				int cCount = 0;
				while (cCount < 10 && ! pq.empty())
				{
					cCount++;
					Bucket b = pq.top(); pq.pop();
					cout << T << " " << cQuery << " " << b.m_low[0] << " " << b.m_low[1] << " " << b.m_count << endl;
				}
			}
		}

		te = time(NULL);
		queryTime += te - ts;

		T++;
	}

	time_t tend = time(NULL);
	time_t totalTime = tend - tstart;

	cerr << "Total Time: " << totalTime << endl;
	cerr << "Query Time: " << queryTime << endl;
	cerr << "Update Time: " << updateTime << endl;

	cout << "Info: " << dataset.size() << " " << gridsize << " " << density << " " << horizon << " " << totalTime << " " << queryTime << " "
		<< updateTime << " " << (gridsize * gridsize * horizon) << endl;

	delete[] grid;
}
