// On-Line Discovery of Dense Areas in Spatio-temporal Databases
// Copyright (C) 2003 Marios Hadjieleftheriou
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Contact information:
//  Mailing address:
//    Marios Hadjieleftheriou
//    University of California, Riverside
//    Department of Computer Science
//    Surge Building, Room 310
//    Riverside, CA 92521
//
//  Email:
//    marioh@cs.ucr.edu

#include "global.h"

class MyBucket : public Bucket
{
	public:
		int m_delta;
};

int main(int argc, char** argv)
{
	char dataFile[1000] = "d";
	int gridsize = 250;
	unsigned int horizon = 20;
	unsigned int density = 1000;
	unsigned int w = 10000;
	bool sampling = false;
	double samplingFactor = 0.01;
	//double epsilon = 0.0;
	int queryType = SNAPSHOT;

	time_t queryTime = 0;
	time_t updateTime = 0;

	for (int i = 1; i < argc; i++)
	{
		if (! strcmp(argv[i], "-df"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing data file." << endl;
				return -1;
			}
			strncpy(dataFile, argv[i], 1000);
		}
		else if (! strcmp(argv[i], "-w"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing w." << endl;
				return -1;
			}
			w = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-gs"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing grid size." << endl;
				return -1;
			}
			gridsize = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-sf"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing sampling factor." << endl;
				return -1;
			}
			sampling = true;
			samplingFactor = atof(argv[i]);
		}
		else if (! strcmp(argv[i], "-hz"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing horizon." << endl;
				return -1;
			}
			horizon = atoi(argv[i]);
		}
		else if (! strcmp(argv[i], "-d"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing density." << endl;
				return -1;
			}
			density = atoi(argv[i]);
		}
		//    else if (! strcmp(argv[i], "-e"))
		//    {
		//      i++;
		//      if (i >= argc)
		//      {
		//        cerr << "Missing epsilon." << endl;
		//        return -1;
		//      }
		//      epsilon = atof(argv[i]);
		//    }
		else if (! strcmp(argv[i], "-qt"))
		{
			i++;
			if (i >= argc)
			{
				cerr << "Missing query type." << endl;
				return -1;
			}
			queryType = (strcasecmp(argv[i], "range") == 0) ? RANGE : SNAPSHOT;
		}
		else
		{
			cerr << "Usage: " << endl
				<< "  -df data file" << endl
				<< "  -gs gridsize" << endl
				<< "  -d density" << endl
				<< "  -w bucket width" << endl
				<< "  -sf sampling factor" << endl
			//           << "  -e epsilon" << endl
				<< "  -hz horizon" << endl
				<< "  -qt query type (snapshot | range)" << endl;
			return -1;
		}
	}

	initRandom();

	int T = TMIN;
	unsigned int currentBucket = 1;
	unsigned long N = 0;
	double epsilon = 1.0 / (double) w;
	int t, ft, id, op, type, t1, t2, iun;
	float x, y, vx, vy, x1, x2, y1, y2;
	char file_buf[1000];
	map<int, MyBucket>** estimator = new map<int, MyBucket>*[horizon];
	map<int, MovingObject> dataset;
	map<int, MovingObject> sample;
	vector<Query> queries;
	double stepx = (XMAX - XMIN) / (double) gridsize;
	double stepy = (YMAX - YMIN) / (double) gridsize;
	unsigned long estimatorSize = 0;
	unsigned long maxEstimatorSize = 0;
	double averageEstimatorSize = 0.0;
	double averageEstimatorSizeCount = 0.0;

	// 1 / epsilon = a * N * u * h, w = u * a * N, u in [0, 1]
	//unsigned int w = (unsigned int) ((10.0 * (double) datasetSize) / ((double) horizon * (double) density));

	for (unsigned int cIndex = 0; cIndex < horizon; cIndex++)
	{
		estimator[cIndex] = new map<int, MyBucket>();
	}

	FILE* f = fopen(dataFile, "r");
	if (f == NULL)
	{
		cerr << "Cannot open " << dataFile << "." << endl;
		return -1;
	}

	cerr << "Data file: " << dataFile << ", gridsize: " << gridsize << ", density: " << density
		<< ", stepx: " << stepx << ", stepy: " << stepy << ", horizon: " << horizon << ", w: " << w << ", epsilon: " << epsilon << endl;

	time_t tstart = time(NULL);

	fgets(file_buf, 1000, f);

	int H = TMIN + horizon;

	while (! feof(f))
	{
		queries.clear();

		time_t ts = time(NULL);

		if (T >= H)
		{
			// rebuild the estimator. Copy paste here for now.
			H += horizon;
			estimatorSize = 0;
			currentBucket = 1;
			N = 0;

			for (unsigned int cTime = 0; cTime < horizon; cTime++)
			{
				estimator[cTime]->clear();
			}

			map<int, MovingObject>* m;
			if (sampling) m = &sample;
			else m = &dataset;

			for (map<int, MovingObject>::iterator it = m->begin(); it != m->end(); it++)
			{
				MovingObject& o = (*it).second;

				for (unsigned int cTime = 0; cTime < horizon; cTime++)
				{
					int cX = (int) ((o.getX(H - horizon + cTime) - XMIN) / stepx);
					int cY = (int) ((o.getY(H - horizon + cTime) - YMIN) / stepy);

					if (cX >= 0 && cX < gridsize && cY >= 0 && cY < gridsize)
					{
						int cellID = (cX * gridsize) + cY;

						if (N > 0 && N % w == 0)
						{
							for (unsigned int cIndex = 0; cIndex < horizon; cIndex++)
							{
								stack<int> del;
								for (map<int, MyBucket>::iterator it2 = estimator[cIndex]->begin(); it2 != estimator[cIndex]->end(); it2++)
								{
									if ((*it2).second.m_count + (*it2).second.m_delta <= currentBucket) del.push((*it2).first);
								}

								while (! del.empty())
								{
									estimator[cIndex]->erase(del.top());
									estimatorSize--;
									del.pop();
								}
							}

							currentBucket++;
						}

						N += 1;

						map<int, MyBucket>::iterator it3 = estimator[cTime]->find(cellID);
						if (it3 != estimator[cTime]->end())
						{
							(*it3).second.m_count++;
						}
						else
						{
							MyBucket b;
							b.m_low[0] = cX;
							b.m_high[0] = cX;
							b.m_low[1] = cY;
							b.m_high[1] = cY;
							b.m_count = 1;
							b.m_delta = currentBucket - 1;
							estimator[cTime]->insert(pair<int, MyBucket>(cellID, b));
							estimatorSize++;
							maxEstimatorSize = max(maxEstimatorSize, estimatorSize);
						}
					}
				}
			}
		}

		while (! feof(f))
		{
			sscanf(file_buf, "%d %d %d", &id, &op, &t);

			if (t > T) break;

			if (op == DELETE)
			{
				sscanf(file_buf, "%d %d %d %d %d %f %f %d %f %f", &id, &op, &t, &ft, &iun, &x, &vx, &iun, &y, &vy);
				MovingObject o(id, x, y, vx, vy, ft);
				dataset.erase(id);

				for (unsigned int cTime = T - (H - horizon); cTime < horizon; cTime++)
				{
					int cX = (int) ((o.getX(H - horizon + cTime) - XMIN) / stepx);
					int cY = (int) ((o.getY(H - horizon + cTime) - YMIN) / stepy);

					if (cX >= 0 && cX < gridsize && cY >= 0 && cY < gridsize)
					{
						int cellID = (cX * gridsize) + cY;

						map<int, MyBucket>::iterator it = estimator[cTime]->find(cellID);
						if (it != estimator[cTime]->end())
						{
							(*it).second.m_count--;
							//            	if ((*it).second.m_count == 0)
							//            	{
							//              	estimator[cTime]->erase(it);
							//	              estimatorSize--;
							//  	          }
						}
					}
				}
			}
			else if (op == QUERY)
			{
				// store queries for current time.
				sscanf(file_buf, "%d %d %d %d %f %f %f %f %d %d %d", &id, &op, &t, &ft, &x1, &x2, &y1, &y2, &t1, &t2, &type);
				Query q(x1, x2, y1, y2, t1, t2);
				queries.push_back(q);
			}
			else if (op == INSERT)
			{
				sscanf(file_buf, "%d %d %d %d %d %f %f %d %f %f", &id, &op, &t, &ft, &iun, &x, &vx, &iun, &y, &vy);
				MovingObject o(id, x, y, vx, vy, ft);
				dataset.erase(id);
				dataset.insert(pair<int, MovingObject>(id, o));

				if (sampling)
				{
					if (T == 0)
					{
						if (nextUniformDouble() <= samplingFactor)
						{
							sample.erase(id);
							sample.insert(pair<int, MovingObject>(id, o));
						}
					}
					else
					{
						map<int, MovingObject>::iterator it = sample.find(o.m_id);
						if (it != sample.end())
						{
							sample.erase(id);
							sample.insert(pair<int, MovingObject>(id, o));
						}
					}
				}

				for (unsigned int cTime = T - (H - horizon); cTime < horizon; cTime++)
				{
					int cX = (int) ((o.getX(H - horizon + cTime) - XMIN) / stepx);
					int cY = (int) ((o.getY(H - horizon + cTime) - YMIN) / stepy);

					if (cX >= 0 && cX < gridsize && cY >= 0 && cY < gridsize)
					{
						int cellID = (cX * gridsize) + cY;

						if (N > 0 && N % w == 0)
						{
							for (unsigned int cIndex = 0; cIndex < horizon; cIndex++)
							{
								stack<int> del;
								for (map<int, MyBucket>::iterator it = estimator[cIndex]->begin(); it != estimator[cIndex]->end(); it++)
								{
									if ((*it).second.m_count + (*it).second.m_delta <= currentBucket) del.push((*it).first);
								}

								while (! del.empty())
								{
									estimator[cIndex]->erase(del.top());
									estimatorSize--;
									del.pop();
								}
							}

							currentBucket++;
						}

						N += 1;

						map<int, MyBucket>::iterator it = estimator[cTime]->find(cellID);
						if (it != estimator[cTime]->end())
						{
							(*it).second.m_count++;
						}
						else
						{
							MyBucket b;
							b.m_low[0] = cX;
							b.m_high[0] = cX;
							b.m_low[1] = cY;
							b.m_high[1] = cY;
							b.m_count = 1;
							b.m_delta = currentBucket - 1;
							estimator[cTime]->insert(pair<int, MyBucket>(cellID, b));
							estimatorSize++;
							maxEstimatorSize = max(maxEstimatorSize, estimatorSize);
						}
					}
				}
			}

			fgets(file_buf, 1000, f);
		}

		time_t te = time(NULL);
		updateTime += te - ts;

		averageEstimatorSize += estimatorSize;
		averageEstimatorSizeCount += 1.0;

		cerr << "Time: " << T << ", Queries: " << queries.size() << endl;

		//    for (unsigned int cTime = 0; cTime < horizon; cTime++)
		//    {
		//      cerr << estimator[cTime]->size() << endl;
		//    }

		ts= time(NULL);

		if (queryType == SNAPSHOT)
		{
			for (unsigned int cQuery = 0; cQuery < queries.size(); cQuery++)
			{
				if (queries[cQuery].m_t1 >= H)
				{
					cout << "ERROR. Time: " << T << ", Query: " << cQuery << " out of estimator horizon. " << queries[cQuery].m_t1 << endl;
					continue;
				}

				//int f = (int) (((double) density) - (epsilon * (double) density));
				int f = (int) (((double) density) - (epsilon * (double) N));

				for (map<int, MyBucket>::iterator it = estimator[queries[cQuery].m_t1 - (H - horizon)]->begin(); it != estimator[queries[cQuery].m_t1 - (H - horizon)]->end(); it++)
				{
					if ((int) (*it).second.m_count >= f)
					{
						cout << T << " " << cQuery << " " << (*it).second.m_low[0] << " " << (*it).second.m_low[1] << " " << (*it).second.m_count << " " << queries[cQuery].m_t1 << endl;
					}
				}
			}
		}
		else
		{
			int f = (int) (((double) density) - (epsilon * (double) N));
			set<int> found[horizon];
			for (unsigned int cTime = T - (H - horizon); cTime < horizon; cTime++)
			{
				for (map<int, MyBucket>::iterator it = estimator[cTime]->begin(); it != estimator[cTime]->end(); it++)
				{
					if ((int) (*it).second.m_count < f) continue;

					int cellID = (int) (*it).second.m_low[0] * gridsize + (int) (*it).second.m_low[1];
					set<int>::iterator itFound = found[cTime].find(cellID);
					if (itFound != found[cTime].end()) continue;

					int interval = 1;
					for (unsigned int cNext = cTime + 1; cNext < horizon; cNext++)
					{
						map<int, MyBucket>::iterator it2 = estimator[cNext]->find(cellID);
						if (it2 != estimator[cNext]->end() && (int) (*it2).second.m_count >= f) interval++;
						else break;
					}
					found[cTime].insert(cellID);
					cout << T << " " << (*it).second.m_low[0] << " " << (*it).second.m_low[1] << " " << (*it).second.m_count << " " << interval << endl;
				}
			}
		}

		te = time(NULL);
		queryTime += te - ts;

		T++;
	}

	time_t tend = time(NULL);
	time_t totalTime = tend - tstart;

	cerr << "Total Time: " << totalTime << endl;
	cerr << "Query Time: " << queryTime << endl;
	cerr << "Update Time: " << updateTime << endl;
	cerr << "Average estimator size per time instant: " << (averageEstimatorSize / averageEstimatorSizeCount) * 2.0 << endl;
	cerr << "Maximum estimator size: " << maxEstimatorSize << endl;

	cout << "Info: " << dataset.size() << " " << gridsize << " " << density << " " << horizon << " " << totalTime << " " << queryTime << " "
		<< updateTime << " " << (averageEstimatorSize / averageEstimatorSizeCount) * 2.0 << endl;
}
